import std.stdio;
import std.stream;
import std.string;
import std.conv;
import std.regex;
import std.file;

struct BitmapFileHeader
{
  ushort bfType ;
  uint   bfSize ;
  ushort bfReserved1 ;
  ushort bfReserved2 ;
  uint   bfOffBits ;
}

struct CIEXYZ
{
  uint ciexyzX ;
  uint ciexyzY ;
  uint ciexyzZ ;
}

struct CIEXYZTRIPLE
{
    CIEXYZ ciexyzRed;
    CIEXYZ ciexyzGreen;
    CIEXYZ ciexyzBlue;
}

enum {
  BI_RGB       =0 ,
  BI_RLE8      =1 ,
  BI_RLE4      =2 ,
  BI_BITFIELDS =3 ,
}

struct BitmapInformationHeader
{
  uint   Size ;
  int    Width ;
  int    Height ;
  ushort Planes ;
  ushort BitCount ;
  uint   Compression ;
  uint   SizeImage ;
  int    XPixPerMeter ;
  int    YPixPerMeter ;
  uint   ClrUsed ;
  uint   CirImportant ;

  uint   RedMask;
  uint   GreenMask;
  uint   BlueMask;
  uint   AlphaMask;
  uint   CSType;
  CIEXYZTRIPLE Endpoints ;
  uint   GammaRed;
  uint   GammaGreen;
  uint   GammaBlue;
  uint   Intent;
  uint   ProfileData;
  uint   ProfileSize;
  uint   Reserved;
} ;

enum FMTtype{
  UNKNOWN ,
  CORE ,
  INFO ,
  V4 ,
  V5 ,
} ;

class FileBMP
{
  BufferedFile infile ;
  int    width ;
  int    height ;
  int    bpp ;
  int    curline ;
  BitmapFileHeader bmpfh ;
  BitmapInformationHeader bi ;
  FMTtype fmttype ;
  uint[]  pallet ;

  this(){
  }

  ~this(){
  }

  int getWidth(){
    return(width) ;
  }

  int getHeight(){
    return(height) ;
  }

  int getBPP(){
    return(bpp) ;
  }

  int open(string fname, string mode="r"){
    infile = new BufferedFile(fname) ;
    int w,h,b ;
    if( mode == "r" ){
      if( readinfo(&w,&h,&b)!=0 ){
        return(-1) ;
      }
      if( bi.Compression!=BI_RGB && bi.Compression!=BI_BITFIELDS ){
        return(-1) ;
      }
      width =w ;
      height=h ;
      bpp   =b ;
      curline=0 ;
    }
    return(0) ;
  }

  int readinfo(int *w, int *h, int *b){
    *w=0 ;
    *h=0 ;
    *b=0 ;
    fmttype = FMTtype.UNKNOWN ;

    infile.read(bmpfh.bfType) ;
    infile.read(bmpfh.bfSize) ;
    infile.read(bmpfh.bfReserved1) ;
    infile.read(bmpfh.bfReserved2) ;
    infile.read(bmpfh.bfOffBits) ;

    if( ((bmpfh.bfType & 0x00ff)   )== 'B' &&
        ((bmpfh.bfType & 0xff00)>>8)== 'M' ){
      uint infohdsize ;
      infile.read(infohdsize) ;
      if( infohdsize==12  || infohdsize==40  ||
          infohdsize==52  || infohdsize==56  || infohdsize==96 || infohdsize==108 ||
          infohdsize==112 || infohdsize==120 || infohdsize==124 ){
        fmttype = FMTtype.CORE ;
        bi.Size=infohdsize ;
        infile.read(bi.Width) ;
        infile.read(bi.Height) ;
        infile.read(bi.Planes) ;
        infile.read(bi.BitCount) ;
        *w = bi.Width ;
        *h = bi.Height * (bi.Height<0 ? -1 : 1)  ;
        *b = 24 ;
      }else{
        return(-1) ;
      }

      if( infohdsize==40  ||
          infohdsize==52  || infohdsize==56  || infohdsize==96 || infohdsize==108 ||
          infohdsize==112 || infohdsize==120 || infohdsize==124 ){
        fmttype = FMTtype.INFO ;
        infile.read(bi.Compression) ;
        infile.read(bi.SizeImage) ;
        infile.read(bi.XPixPerMeter) ;
        infile.read(bi.YPixPerMeter) ;
        infile.read(bi.ClrUsed) ;
        infile.read(bi.CirImportant) ;
      }
      if( infohdsize==52  || infohdsize==56  || infohdsize==96 || infohdsize==108 ||
          infohdsize==112 || infohdsize==120 || infohdsize==124 ){
        fmttype = FMTtype.V4 ;
        infile.read(bi.RedMask) ;
        infile.read(bi.GreenMask) ;
        infile.read(bi.BlueMask) ;
      }
      if( infohdsize==56  || infohdsize==96 || infohdsize==108 ||
          infohdsize==112 || infohdsize==120 || infohdsize==124 ){
        infile.read(bi.AlphaMask) ;
      }
      if( infohdsize==96 || infohdsize==108 ||
          infohdsize==112 || infohdsize==120 || infohdsize==124 ){
        infile.read(bi.CSType) ;
        infile.read(bi.Endpoints.ciexyzRed.ciexyzX) ;
        infile.read(bi.Endpoints.ciexyzRed.ciexyzY) ;
        infile.read(bi.Endpoints.ciexyzRed.ciexyzZ) ;
        infile.read(bi.Endpoints.ciexyzGreen.ciexyzX) ;
        infile.read(bi.Endpoints.ciexyzGreen.ciexyzY) ;
        infile.read(bi.Endpoints.ciexyzGreen.ciexyzZ) ;
        infile.read(bi.Endpoints.ciexyzBlue.ciexyzX) ;
        infile.read(bi.Endpoints.ciexyzBlue.ciexyzY) ;
        infile.read(bi.Endpoints.ciexyzBlue.ciexyzZ) ;
      }
      if( infohdsize==108 ||
          infohdsize==112 || infohdsize==120 || infohdsize==124 ){
        infile.read(bi.GammaRed) ;
        infile.read(bi.GammaGreen) ;
        infile.read(bi.GammaBlue) ;
      }
      if( infohdsize==112 || infohdsize==120 || infohdsize==124 ){
        fmttype = FMTtype.V5 ;
        infile.read(bi.Intent) ;
      }
      if( infohdsize==120 || infohdsize==124 ){
        infile.read(bi.ProfileData) ;
        infile.read(bi.ProfileSize) ;
      }
      if( infohdsize==124 ){
        infile.read(bi.Reserved) ;
      }
    }else{
      return(-1) ;
    }

    if( bi.BitCount<=8 ){
      pallet.length= (1<<bi.BitCount) ;
      if( fmttype==FMTtype.CORE ){
        for( int i=0 ; i<pallet.length ; i++ ){
          infile.read(pallet[i]) ;
        }
      }else{
        for( int i=0 ; i<pallet.length ; i++ ){
          ubyte R,G,B,X ;
          infile.read(B) ; infile.read(G) ; infile.read(R) ; infile.read(X) ;
          pallet[i]=cast(uint)(B)<<16 | cast(uint)(G)<<8 | cast(uint)(R) ;
        }
      }
    }

    return(0) ;
  }

  uint[] readline(int lnum=1){
    uint[]  img ;
    char[]  rbuf ;

    img.length=width*lnum ;
    if( fmttype!=FMTtype.UNKNOWN ){
      if( bi.Compression==BI_RGB ){
        int recwidth = (((width * bi.BitCount) +31)/32)*4 ;
        for( int y=0 ; y<lnum ; y++ ){
          if( bi.Height>=0 ){
            infile.seekSet( bmpfh.bfOffBits + (height - curline - 1)*recwidth ) ;
          }
          rbuf = infile.readString(recwidth) ;
          switch( bi.BitCount ){
          case 1 :
            for( int x=0 ; x<width ; x++ ){
              int bytepos = x/8 ;
              int bitpos  = (7 - x%8) ;
              img[width*y+x]=0xff000000 | pallet[(rbuf[bytepos] >> bitpos) & 0x01] ;
            }
            break ;
          case 4 :
            for( int x=0 ; x<width ; x++ ){
              int bytepos = x/2 ;
              int bitpos  = (1-(x%2))*4 ;
              img[width*y+x]=0xff000000 | pallet[(rbuf[bytepos] >> bitpos) & 0x0f] ;
            }
            break ;
          case 8 :
            for( int x=0 ; x<width ; x++ ){
              img[width*y+x]=0xff000000 | pallet[rbuf[x]] ;
            }
            break ;
          case 24 :
            for( int x=0 ; x<width ; x++ ){
              img[width*y+x]=0xff000000 |
                             (rbuf[x*3+0]<<16) |
                             (rbuf[x*3+1]<< 8) |
                             (rbuf[x*3+2]    ) ;
            }
            break ;
          case 32 :
            for( int x=0 ; x<width ; x++ ){
              img[width*y+x]=0xff000000 |
                             (rbuf[x*4+0]<<16) |
                             (rbuf[x*4+1]<< 8) |
                             (rbuf[x*4+2]    ) ;
            }
            break ;
          default :
            break ;
          }
          curline++ ;
        }
      }

      if( bi.Compression==BI_BITFIELDS ){
        int recwidth = (((width * bi.BitCount) +31)/32)*4 ;
        for( int y=0 ; y<lnum ; y++ ){
          if( bi.Height>=0 ){
            infile.seekSet( bmpfh.bfOffBits + (height - curline - 1)*recwidth ) ;
          }
          rbuf = infile.readString(recwidth) ;
          switch( bi.BitCount ){
          case 16 :
            if(       bi.RedMask==0x7c00 && bi.GreenMask==0x03e00 && bi.BlueMask==0x001f ){
              for( int x=0 ; x<width ; x++ ){
                ushort pix = rbuf[x*2+0] | (rbuf[x*2+1]<<8) ;
                img[width*y+x]=0xff000000 |
                               ((pix & 0x7c00)>>8) |
                               ((pix & 0x03e0)>>2) |
                               ((pix & 0x001f)<<3) ;
              }
            }else if( bi.RedMask==0xf800 && bi.GreenMask==0x07e00 && bi.BlueMask==0x001f ){
              for( int x=0 ; x<width ; x++ ){
                ushort pix = rbuf[x*2+0] | (rbuf[x*2+1]<<8) ;
                img[width*y+x]=0xff000000 |
                               ((pix & 0xf800)>>8) |
                               ((pix & 0x03e0)>>3) |
                               ((pix & 0x001f)<<3) ;
              }
            }
            break ;
          case 32 :
            if( bi.RedMask==0xff000000 && bi.GreenMask==0x00ff0000 && bi.BlueMask==0x0000ff00 ){
              for( int x=0 ; x<width ; x++ ){
                uint pix = rbuf[x*4+0] | (rbuf[x*4+1]<<8) | (rbuf[x*4+2]<<16) | (rbuf[x*4+3]<<24) ;
                img[width*y+x]=0xff000000 |
                               ((pix & 0xff000000)>>8) |
                               ((pix & 0x00ff0000)>>8) |
                               ((pix & 0x0000ff00)>>8) ;
              }
            }
            else if( bi.RedMask==0x00ff0000 && bi.GreenMask==0x0000ff00 && bi.BlueMask==0x000000ff ){
              for( int x=0 ; x<width ; x++ ){
                uint pix = rbuf[x*4+0] | (rbuf[x*4+1]<<8) | (rbuf[x*4+2]<<16) | (rbuf[x*4+3]<<24) ;
                img[width*y+x]=0xff000000 |
                               (pix & 0x00ff0000) |
                               (pix & 0x0000ff00) |
                               (pix & 0x000000ff) ;
              }
            }
            break ;
          default :
            break ;
          }
          curline++ ;
        }
      }
    }
    return(img) ;
  }

  int close(){
    if( infile !is null ){
      infile.close() ;
    }
    return(0) ;
  }
}

version (BMP_DEBUG){
int main(string[] args)
{
  FileBMP bmp = new FileBMP() ;
  bmp.open(args[1]) ;
  writef("w=%d h=%d b=%d\n", bmp.width, bmp.height, bmp.bpp) ;
  int w = bmp.getWidth() ;
  int h = bmp.getHeight() ;

  BufferedFile outfile ;
  char[]       cimg ;
  outfile = new BufferedFile() ;
  outfile.create("outputtest.ppm") ;
  outfile.writef("P6\n%d %d\n255\n",w,h) ;

  cimg.length=w*h*3 ;
  uint[] l ;
  l = bmp.readline(h) ;
  for( int x=0 ; x<w*h ; x++ ){
    cimg[x*3+2] = (l[x] & 0x00ff0000)>>16 ;
    cimg[x*3+1] = (l[x] & 0x0000ff00)>>8  ;
    cimg[x*3+0] = (l[x] & 0x000000ff)     ;
  }
  outfile.writeString(cimg) ;

  outfile.close() ;

  return(0) ;
}
}
