import std.stdio;
import std.stream;
import std.string;
import std.conv;
import std.regex;
import std.file;

enum {
  BI_RGB       =0 ,
  BI_RLE8      =1 ,
  BI_RLE4      =2 ,
  BI_BITFIELDS =3 ,
}

struct DDSHeader
{
  uint     Magic ;
  uint     Size ;
  uint     Flags ;
  uint     Height ;
  uint     Width ;
  uint     PitchOrLinearSize ;

  uint     Depth ;
  uint     MipMapCount ;
  uint[11] Reserved1 ;

  uint     PfSize ;
  uint     PfFlags ;
  uint     FourCC ;

  uint     RGBBitCount ;
  uint     RBitMask ;
  uint     GBitMask ;
  uint     BBitMask ;
  uint     RGBAlphaBitMask ;
  uint     Caps ;
  uint     Caps2 ;
  uint[2]  ReservedCaps ;
  uint     Reserved2 ;
} ;

struct DDSHeader_DX10
{
  uint Format ;
  uint Dimension ;
  uint MiscFlag ;
  uint ArraySize ;
  uint MiscFlag2 ;
}

uint FOURCC( ubyte x, ubyte y, ubyte z, ubyte w ) {
  return( (w<<24)|(z<<16)|(y<<8)|(x) ) ;
}

class FileDDS
{
  BufferedFile infile ;
  int    width ;
  int    height ;
  int    bpp ;
  int    curline ;
  DDSHeader ddshd ;
  DDSHeader_DX10 ddsexhd ;

  uint[] pixbuf ;
  int    bufptr=4 ;
  int    blkcount=0 ;

  this(){
  }

  ~this(){
  }

  int getWidth(){
    return(width) ;
  }

  int getHeight(){
    return(height) ;
  }

  int getBPP(){
    return(bpp) ;
  }

  int open(string fname, string mode="r"){
    infile = new BufferedFile(fname) ;
    int w,h,b ;
    if( mode == "r" ){
      if( readinfo(&w,&h,&b)!=0 ){
        return(-1) ;
      }
      width =w ;
      height=h ;
      bpp   =b ;
      curline=0 ;
      bufptr=4 ;
    }
    return(0) ;
  }

  int readinfo(int *w, int *h, int *b){
    *w=0 ;
    *h=0 ;
    *b=0 ;

    infile.read(ddshd.Magic) ;

    if( ddshd.Magic == FOURCC('D', 'D', 'S', ' ') ){
      infile.read(ddshd.Size) ;
      infile.read(ddshd.Flags) ;
      infile.read(ddshd.Height) ;
      infile.read(ddshd.Width) ;
      infile.read(ddshd.PitchOrLinearSize) ;
      infile.read(ddshd.Depth) ;
      infile.read(ddshd.MipMapCount) ;
      infile.read(ddshd.Reserved1[0]) ;
      infile.read(ddshd.Reserved1[1]) ;
      infile.read(ddshd.Reserved1[2]) ;
      infile.read(ddshd.Reserved1[3]) ;
      infile.read(ddshd.Reserved1[4]) ;
      infile.read(ddshd.Reserved1[5]) ;
      infile.read(ddshd.Reserved1[6]) ;
      infile.read(ddshd.Reserved1[7]) ;
      infile.read(ddshd.Reserved1[8]) ;
      infile.read(ddshd.Reserved1[9]) ;
      infile.read(ddshd.Reserved1[10]) ;
      infile.read(ddshd.PfSize) ;
      infile.read(ddshd.PfFlags) ;
      infile.read(ddshd.FourCC) ;
      infile.read(ddshd.RGBBitCount) ;
      infile.read(ddshd.RBitMask) ;
      infile.read(ddshd.GBitMask) ;
      infile.read(ddshd.BBitMask) ;
      infile.read(ddshd.RGBAlphaBitMask) ;
      infile.read(ddshd.Caps) ;
      infile.read(ddshd.Caps2) ;
      infile.read(ddshd.ReservedCaps[0]) ;
      infile.read(ddshd.ReservedCaps[1]) ;
      infile.read(ddshd.Reserved2) ;
      if( ddshd.FourCC == FOURCC('D', 'X', '1', '0') ){
        infile.read(ddsexhd.Format) ;
        infile.read(ddsexhd.Dimension) ;
        infile.read(ddsexhd.MiscFlag) ;
        infile.read(ddsexhd.ArraySize) ;
        infile.read(ddsexhd.MiscFlag2) ;
      }

    }else{
      return(-1) ;
    }

//    writef("FourCC %d %d %d %d\n",
//           (ddshd.FourCC & 0x000000ff),
//           (ddshd.FourCC & 0x0000ff00)>>8,
//           (ddshd.FourCC & 0x00ff0000)>>16,
//           (ddshd.FourCC & 0xff000000)>>24
//           ) ;
//    writef("bbp %d\n",ddshd.RGBBitCount) ;
//    writef("mipmap %d\n",ddshd.MipMapCount) ;
//    writef("flags %08x\n",ddshd.Flags) ;

    *w=ddshd.Width ;
    *h=ddshd.Height ;
    *b=32 ;

    pixbuf.length =  ddshd.Width*4 ;

    return(0) ;
  }

  uint[] decodeDTXline(){
    if( bufptr>3 ){
      for( int i=0 ; i<width/4 ; i++ ){
//        ubyte    a0, a1 ;
//        ubyte[6] aidx ;
//        int[8]   atbl ;
//
//        infile.read(a0); infile.read(a1);
//        infile.read(aidx[0]); infile.read(aidx[1]); infile.read(aidx[2]);
//        infile.read(aidx[3]); infile.read(aidx[4]); infile.read(aidx[5]);
//
//        atbl[0]=a0 ;
//        atbl[1]=a1 ;
//        if( a0>a1 ){
//          atbl[2] = (6*a0 + 1*a1)/7 ;
//          atbl[3] = (5*a0 + 2*a1)/7 ;
//          atbl[4] = (4*a0 + 3*a1)/7 ;
//          atbl[5] = (3*a0 + 4*a1)/7 ;
//          atbl[6] = (2*a0 + 5*a1)/7 ;
//          atbl[7] = (1*a0 + 6*a1)/7 ;
//        }else{
//          atbl[2] = (4*a0 + 1*a1)/5 ;
//          atbl[3] = (3*a0 + 2*a1)/5 ;
//          atbl[4] = (2*a0 + 3*a1)/5 ;
//          atbl[5] = (1*a0 + 4*a1)/5 ;
//          atbl[6] = 0 ;
//          atbl[7] = 255 ;
//        }

        ushort[4] alpha ;
        infile.read(alpha[0]); infile.read(alpha[1]); infile.read(alpha[2]); infile.read(alpha[3]);
        uint[16]  a ;
        for( int j=0 ; j<4 ; j++ ){
          a[j*4+0] = cast(uint)((alpha[j] & 0x000f))<<28 | cast(uint)((alpha[j] & 0x000f))<<24 ;
          a[j*4+1] = cast(uint)((alpha[j] & 0x00f0))<<24 | cast(uint)((alpha[j] & 0x00f0))<<20 ;
          a[j*4+2] = cast(uint)((alpha[j] & 0x0f00))<<20 | cast(uint)((alpha[j] & 0x0f00))<<16 ;
          a[j*4+3] = cast(uint)((alpha[j] & 0xf000))<<16 | cast(uint)((alpha[j] & 0xf000))<<12 ;
        }

        ushort   c0, c1 ;
        ubyte[4] cidx ;
        infile.read(c0); infile.read(c1);
        infile.read(cidx[0]); infile.read(cidx[1]); infile.read(cidx[2]); infile.read(cidx[3]);

        uint[4]  ctbl ;
        int      c0r,c0g,c0b,c1r,c1g,c1b ;

        c0r=(c0 >> 11) & 0x1f ; c0r = (c0r<<3) | (c0r>>2) ;
        c0g=(c0 >>  5) & 0x3f ; c0g = (c0g<<2) | (c0g>>4) ;
        c0b=(c0      ) & 0x1f ; c0b = (c0b<<3) | (c0b>>2) ;
        c1r=(c1 >> 11) & 0x1f ; c1r = (c1r<<3) | (c1r>>2) ;
        c1g=(c1 >>  5) & 0x3f ; c1g = (c1g<<2) | (c1g>>4) ;
        c1b=(c1      ) & 0x1f ; c1b = (c1b<<3) | (c1b>>2) ;

        ctbl[0] = c0b<<16 | c0g<<8 | c0r ;
        ctbl[1] = c1b<<16 | c1g<<8 | c1r ;
        ctbl[2] =((c0b*2 + c1b*1)/3)<<16 | ((c0g*2 + c1g*1)/3)<<8 | ((c0r*2 + c1r*1)/3) ;
        ctbl[3] =((c0b*1 + c1b*2)/3)<<16 | ((c0g*1 + c1g*2)/3)<<8 | ((c0r*1 + c1r*2)/3) ;

        for( int j=0 ; j<4 ; j++ ){
          pixbuf[width*j+i*4+3] = ctbl[ (cidx[j]&0xc0)>>6 ] | a[j*4+3] ;
          pixbuf[width*j+i*4+2] = ctbl[ (cidx[j]&0x30)>>4 ] | a[j*4+2] ;
          pixbuf[width*j+i*4+1] = ctbl[ (cidx[j]&0x0c)>>2 ] | a[j*4+1] ;
          pixbuf[width*j+i*4+0] = ctbl[ (cidx[j]&0x03)    ] | a[j*4+0] ;
        }
      }
      bufptr=0 ;
    }

    uint[]  imgline ;
    imgline.length=width ;
    for( int i=0 ; i<width ; i++ ){
      imgline[i]=pixbuf[bufptr*width+i] ;
    }
    bufptr++ ;
    return(imgline) ;
  }


  uint[] readline(int lnum=1){
    uint[] img ;
    img.length=width*lnum ;

    if( curline<height ){
      for( int y=0 ; y<lnum ; y++ ){
        uint[] line = decodeDTXline() ;
        for( int x=0 ; x<width ; x++ ){
          img[y*width+x]=line[x] ;
        }
        curline++ ;
        if( curline>=height ) break ;
      }
    }
    return(img) ;
  }

  int close(){
    if( infile !is null ){
      infile.close() ;
    }
    return(0) ;
  }
}

version (DDS_DEBUG){
int main(string[] args)
{
  FileDDS dds = new FileDDS() ;
  dds.open(args[1]) ;
  writef("w=%d h=%d b=%d\n", dds.width, dds.height, dds.bpp) ;
  int w = dds.getWidth() ;
  int h = dds.getHeight() ;

  BufferedFile outfile ;
  char[]       cimg ;
  outfile = new BufferedFile() ;
  outfile.create("outputtest.ppm") ;
  outfile.writef("P6\n%d %d\n255\n",w,h) ;

  cimg.length=w*h*3 ;
  uint[] l ;
  for( int y=0 ; y<h ; y++ ){
    l = dds.readline(1) ;
    for( int x=0 ; x<w ; x++ ){
      int a=(l[x] & 0xff000000)>>24 ;
      cimg[(w*3)*y+x*3+2] = cast(char)(cast(int)((l[x] & 0x00ff0000)>>16)*a/255) ;
      cimg[(w*3)*y+x*3+1] = cast(char)(cast(int)((l[x] & 0x0000ff00)>>8 )*a/255) ;
      cimg[(w*3)*y+x*3+0] = cast(char)(cast(int)((l[x] & 0x000000ff)    )*a/255) ;
    }
  }
  outfile.writeString(cimg) ;

  outfile.close() ;

  return(0) ;
}
}
