import std.stdio;
import std.stream;
import std.string;
import std.conv;
import std.regex;
import std.file;

struct TGAHEADER
{
  ubyte  IDlength ;
  ubyte  colormaptype ;
  ubyte  imagetype ;
  ushort colormaporigin ;
  ushort colormaplength ;
  ubyte  colormapdepth ;
  ushort originX ;
  ushort originY ;
  ushort width ;
  ushort height ;
  ubyte  depth ;
  ubyte  descriptor ;
}

enum {
  NONE         = 0,
  COLORMAP     = 1,
  RGB          = 2,
  GRAYSCALE    = 3,
  RLE_COLORMAP = 9,
  RLE_RGB      =10,
  RLE_GRAYSCALE=11,
}

class FileTGA
{
  BufferedFile infile ;
  TGAHEADER header ;
  int      width ;
  int      height ;
  int      bpp ;
  int      curline ;
  int      headersize=18;
  uint[]   colmap ;
  uint[][] imgbuf ;

  this(){
  }

  ~this(){
    for( int i=0 ; i<imgbuf.length ; i++ ){
      delete imgbuf[i] ;
    }
    delete imgbuf ;
  }

  int getWidth(){
    return(width) ;
  }

  int getHeight(){
    return(height) ;
  }

  int getBPP(){
    return(bpp) ;
  }

  int open(string fname, string mode="r"){
    infile = new BufferedFile(fname) ;
    int w,h,b ;
    if( mode == "r" ){
      if( readinfo(&w,&h,&b)!=0 ){
        return(-1) ;
      }
      width =w ;
      height=h ;
      bpp   =b ;
      curline=0 ;
//      writef("header=%s\n",header) ;
    }
    return(0) ;
  }

  int readinfo(int *w, int *h, int *b){
    infile.read(header.IDlength) ;
    infile.read(header.colormaptype) ;
    infile.read(header.imagetype) ;
    infile.read(header.colormaporigin) ;
    infile.read(header.colormaplength) ;
    infile.read(header.colormapdepth) ;
    infile.read(header.originX) ;
    infile.read(header.originY) ;
    infile.read(header.width) ;
    infile.read(header.height) ;
    infile.read(header.depth) ;
    infile.read(header.descriptor) ;
    *w = header.width ;
    *h = header.height ;
    *b = header.depth ;

    colmap.length = header.colormaplength ;
    for( int i=0 ; i<colmap.length ; i++ ){
      switch( header.colormapdepth ){
      case 24:
        ubyte red,green,blue ;
        infile.read(red); infile.read(green); infile.read(blue);
        colmap[i] = 0xff000000 | red<<16 | green<<8 | blue ;
        break ;
      case 32 :
        ubyte red,green,blue,alpha ;
        infile.read(alpha); infile.read(red); infile.read(green); infile.read(blue);
        colmap[i] = alpha<<24 | red<<16 | green<<8 | blue ;
        break ;
      default : break ;
      }
    }

    if( header.imagetype==RLE_COLORMAP ||
        header.imagetype==RLE_RGB      ||
        header.imagetype==RLE_GRAYSCALE ){
      int ptr=0 ;
      imgbuf.length = header.height ;
      for( int i=0 ; i<imgbuf.length ; i++ ){
        imgbuf[i].length=header.width ;
      }

      while( ptr < header.height*header.width ){
        ubyte ctl ; infile.read(ctl) ;
        int   count = (ctl & 0x7f) + 1 ;
        if( (ctl & 0x80)!=0 ){
          uint col=readcolor(header.imagetype,header.depth) ;
          for( int i=0 ; i<count ; i++ ){
            imgbuf[ptr/header.width][ptr%header.width]=col ;
            ptr++ ;
          }
        }else{
          for( int i=0 ; i<count ; i++ ){
            imgbuf[ptr/header.width][ptr%header.width]=readcolor(header.imagetype,header.depth) ;
            ptr++ ;
          }
        }
      }
    }
    return(0) ;
  }

  private uint readcolor(int imagetype, int depth){
    uint col ;

    switch( imagetype ){
    case RLE_COLORMAP :
      ubyte idx ;
      infile.read(idx) ;
      col=colmap[idx] ;
      break ;

    case RLE_RGB :
      switch( depth ){
      case 24 :
        ubyte red,green,blue ;
        infile.read(red); infile.read(green); infile.read(blue);
        col = 0xff000000 | red<<16 | green<<8 | blue ;
        break ;
      case 32 :
        ubyte red,green,blue,alpha ;
        infile.read(red); infile.read(green); infile.read(blue); infile.read(alpha);
        col = alpha<<24 | red<<16 | green<<8 | blue ;
        break ;
      default :
        break ;
      }
      break ;

    case RLE_GRAYSCALE :
      switch( depth ){
      case 8 :
        ubyte c ;
        infile.read(c);
        col = 0xff000000 | c<<16 | c<<8 | c ;
        break ;
      case 16 :
        ubyte c,alpha ;
        infile.read(c); infile.read(alpha) ;
        col = alpha<<24 | c<<16 | c<<8 | c ;
        break ;
      default :
        break ;
      }
      break ;

    default :
      break ;
    }

    return(col) ;
  }

  uint[] readline(int lnum=1){
    uint[]  img ;
    char[]  rbuf ;
    img.length=width*lnum ;
    switch( header.imagetype ){
    case RLE_COLORMAP :
    case RLE_RGB :
    case RLE_GRAYSCALE :
      bool dtou= (header.descriptor & 0x20)==0 ;
      bool ltor= (header.descriptor & 0x10)==0 ;
      for( int y=0 ; y<lnum ; y++ ){
        int srcypos ;
        if( dtou==true ){
          srcypos=height - curline - 1 ;
        }else{
          srcypos=curline ;
        }
        for( int x=0 ; x<width ; x++ ){
          img[width*y+x] = imgbuf[srcypos][x] ;
        }
        curline++ ;
      }
      break ;

    case COLORMAP :
      int  colmapsize = header.colormapdepth/8 * header.colormaplength;
      bool dtou= (header.descriptor & 0x20)==0 ;
      bool ltor= (header.descriptor & 0x10)==0 ;
      for( int y=0 ; y<lnum ; y++ ){
        if( dtou==true ){
          infile.seekSet( headersize + colmapsize + (height - curline - 1)*width ) ;
        }
        rbuf = infile.readString(width) ;
        for( int x=0 ; x<width ; x++ ){
          img[width*y+x] = colmap[rbuf[x]] ;
        }
        curline++ ;
      }
      break ;

    case RGB :
      int  Bpp = bpp/8 ;
      bool dtou= (header.descriptor & 0x20)==0 ;
      bool ltor= (header.descriptor & 0x10)==0 ;
      for( int y=0 ; y<lnum ; y++ ){
        if( dtou==true ){
          infile.seekSet( headersize + (height - curline - 1)*width*Bpp ) ;
        }
        rbuf = infile.readString(width*Bpp) ;
        for( int x=0 ; x<width ; x++ ){
          img[width*y+x] = (bpp==32 ? (rbuf[x*Bpp+3]<<24) : 0xff000000) |
                           (rbuf[x*Bpp+0]<<16) |
                           (rbuf[x*Bpp+1]<< 8) |
                           (rbuf[x*Bpp+2]) ;
        }
        curline++ ;
      }
      break ;

    case GRAYSCALE :
      int  Bpp = bpp/8 ;
      bool dtou= (header.descriptor & 0x20)==0 ;
      bool ltor= (header.descriptor & 0x10)==0 ;
      for( int y=0 ; y<lnum ; y++ ){
        if( dtou==true ){
          infile.seekSet( headersize + (height - curline - 1)*width*Bpp ) ;
        }
        rbuf = infile.readString(width*Bpp) ;
        for( int x=0 ; x<width ; x++ ){
          img[width*y+x] = (bpp==16 ? (rbuf[x*Bpp+1]<<24) : 0xff000000) |
                           (rbuf[x*Bpp+0]<<16) |
                           (rbuf[x*Bpp+0]<< 8) |
                           (rbuf[x*Bpp+0]) ;
        }
        curline++ ;
      }
      break ;

    default :
      break ;
    }

    return(img) ;
  }

  int close(){
    if( infile !is null ){
      infile.close() ;
    }
    return(0) ;
  }
}

version (TGA_DEBUG){
int main(string[] args)
{
  FileTGA tga = new FileTGA() ;
  tga.open(args[1]) ;
  writef("w=%d h=%d b=%d\n", tga.width, tga.height, tga.bpp) ;
  int w = tga.getWidth() ;
  int h = tga.getHeight() ;

  BufferedFile outfile ;
  char[]       cimg ;
  outfile = new BufferedFile() ;
  outfile.create("outputtest.ppm") ;
  outfile.writef("P6\n%d %d\n255\n",w,h) ;
//  cimg.length=w*3 ;
//
//  for( int y=0 ; y<h ; y++ ){
//    uint[] l ;
//    l = tga.readline(1) ;
//    for( int x=0 ; x<w ; x++ ){
//      cimg[x*3+2] = (l[x] & 0x00ff0000)>>16 ;
//      cimg[x*3+1] = (l[x] & 0x0000ff00)>>8  ;
//      cimg[x*3+0] = (l[x] & 0x000000ff)     ;
//    }
//    outfile.writeString(cimg) ;
//  }

  cimg.length=w*h*3 ;
  uint[] l ;
  l = tga.readline(h) ;
  for( int x=0 ; x<w*h ; x++ ){
    cimg[x*3+2] = (l[x] & 0x00ff0000)>>16 ;
    cimg[x*3+1] = (l[x] & 0x0000ff00)>>8  ;
    cimg[x*3+0] = (l[x] & 0x000000ff)     ;
  }
  outfile.writeString(cimg) ;

  outfile.close() ;

  return(0) ;
}
}
